package juniq;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import java.io.*;
import java.util.*;
import java.util.concurrent.*;
import java.security.*;
import javax.swing.table.*;

public class JUniq implements Serializable
{
    public static final long serialVersionUID = 1001L;

    JFrame frame;
    JTextArea textarea;

    static final int FASTHASH_LENGTH = 512;
    static final int READBUF_LENGTH = 32768;

    JMenuBar menuBar = new JMenuBar();
    JMenu    fileMenu = new JMenu("Database");
    JMenuItem newItem = new JMenuItem("New");
    JMenuItem openItem = new JMenuItem("Open...");
    JMenuItem saveItem = new JMenuItem("Save");
    JMenuItem saveAsItem = new JMenuItem("Save as...");
    JMenuItem quitItem = new JMenuItem("Quit");

    JMenu     opMenu = new JMenu("Operations");
    JMenuItem addItem = new JMenuItem("Add path...");
    JMenuItem genItem = new JMenuItem("Generate delete script...");
    JMenuItem emptyDirsItem = new JMenuItem("Find empty folders...");

    HashMap<Long, ArrayList<FileSet>> digestTable = new HashMap<Long, ArrayList<FileSet>>();
    ArrayList<FileSet> allFileSets = new ArrayList<FileSet>();
    ArrayList<FileSet> dupFileSets = new ArrayList<FileSet>();
    ArrayList<File>    symLinks = new ArrayList<File>();
    ArrayList<File>    nonFiles = new ArrayList<File>();

    Stats stats = new Stats();

    String    currentDbPath = null;
    JFileChooser  dbChooser = new JFileChooser();
    JFileChooser  pathChooser = new JFileChooser();

    AbstractTableModel tableModel = new UniqTableModel();
    TableSorter tableSorter = new TableSorter(tableModel); //, table.getTableHeader());
    JTable    table = new JTable(tableSorter);

    JList     pathList = new JList();

    JLabel  locatedFilesLabel = new JLabel("Located: 0 Files");
    JLabel  processedFilesLabel = new JLabel("Processed: 0 Files");
    //    JLabel  processedMbLabel = new JLabel("Processed: 0 MB");
    JLabel  readMbLabel = new JLabel("I/O: 0 MB");
    JLabel  discardMbLabel = new JLabel("Discardable: 0 MB");
    JLabel  filesetsLabel = new JLabel("Filesets: 0");

    ArrayList<JLabel> statusLabels = new ArrayList<JLabel>();

    BlockingQueue<File>  fileQueue = new LinkedBlockingQueue<File>();
    BlockingQueue<File>  rootQueue = new LinkedBlockingQueue<File>();

    int numProcessThreads = 1;
    ArrayList<ProcessThread> processThreads = new ArrayList<ProcessThread>();

    Generate generate = new Generate(this);

    public JUniq()
    {
	frame = new JFrame("JUniq 0.4, by Edwin Olson (eolson@mit.edu)");
	frame.getContentPane().setLayout(new BorderLayout());

	textarea = new JTextArea();
	textarea.setEditable(false);
	textarea.setFont(new Font("Monospaced", Font.PLAIN, 10));

	table.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);

	menuBar.add(fileMenu);
	fileMenu.add(newItem);
	fileMenu.add(openItem);
	fileMenu.add(saveItem);
	fileMenu.add(saveAsItem);
	fileMenu.add(quitItem);

	menuBar.add(opMenu);
	opMenu.add(addItem);
	opMenu.add(genItem);
	opMenu.add(emptyDirsItem);

	newItem.addActionListener(new ActionListener() 
	    { public void actionPerformed(ActionEvent e) { fileNew(); }});
	openItem.addActionListener(new ActionListener() 
	    { public void actionPerformed(ActionEvent e) { fileOpen(); }});
	saveItem.addActionListener(new ActionListener() 
	    { public void actionPerformed(ActionEvent e) { fileSave(); }});
	saveAsItem.addActionListener(new ActionListener() 
	    { public void actionPerformed(ActionEvent e) { fileSaveAs(); }});
	quitItem.addActionListener(new ActionListener() 
	    { public void actionPerformed(ActionEvent e) { fileQuit(); }});
	addItem.addActionListener(new ActionListener() 
	    { public void actionPerformed(ActionEvent e) { opAdd(); }});
	genItem.addActionListener(new ActionListener() 
	    { public void actionPerformed(ActionEvent e) { opGenerate(); }});
	emptyDirsItem.addActionListener(new ActionListener()
	    { public void actionPerformed(ActionEvent e) { new EmptyFolders(); }});

	frame.setJMenuBar(menuBar);

	JPanel statsbox = new JPanel(new GridLayout(2,3)); //new GridLayout());
	statsbox.add(locatedFilesLabel);
	statsbox.add(processedFilesLabel);
	statsbox.add(discardMbLabel);
	//	statsbox.add(processedMbLabel);
	statsbox.add(readMbLabel);
	statsbox.add(filesetsLabel);
	statsbox.add(new JLabel(""));

	Box hbox1 = new Box(BoxLayout.Y_AXIS);
	//	hbox1.add(new JLabel("Duplicate file sets:"));
	hbox1.add(new JScrollPane(table));
	//	table.setRowHeight(50);
	//	hbox.add(new JScrollPane(textarea));
	Box hbox2 = new Box(BoxLayout.Y_AXIS);
	//	hbox2.add(new JLabel("Duplicate files in selected set:"));
	hbox2.add(new JScrollPane(pathList));
	hbox2.add(statsbox);

	DirectorySearcherThread dst = new DirectorySearcherThread();
	dst.start();

	Box statuslabelsbox = new Box(BoxLayout.Y_AXIS);
	for (int i=0; i < numProcessThreads; i++)
	    {
		ProcessThread pt = new ProcessThread();
		processThreads.add(pt);
		pt.start();

		JLabel jl  = new JLabel("Idle");
		statuslabelsbox.add(jl);
		statusLabels.add(jl);
	    }

	StatusThread st = new StatusThread();
	st.start();

	/////////////////////////////////////////////////
	JSplitPane jsp = new JSplitPane(JSplitPane.VERTICAL_SPLIT, hbox1, hbox2);
	jsp.setDividerLocation(0.8);
	jsp.setResizeWeight(0.8);

	frame.getContentPane().add(jsp, BorderLayout.CENTER);
	frame.getContentPane().add(statuslabelsbox, BorderLayout.SOUTH);

	fileNew();

	ListSelectionModel rowSM = table.getSelectionModel();
	rowSM.addListSelectionListener(new ListSelectionListener() {
		public void valueChanged(ListSelectionEvent e)
		{
		    tableSelectionChanged();
		}
	    });
	TableColumnModel tcm = table.getColumnModel();
	tcm.getColumn(1).setMaxWidth(80);
	tcm.getColumn(2).setMaxWidth(300);
	tcm.getColumn(3).setMaxWidth(300);
	//	table.setModel(tableSorter);
	tableSorter.setTableHeader(table.getTableHeader());
	//	tableSorter.setColumnComparator(
	frame.setSize(800,600);
	frame.setVisible(true);
    }

    public void tableSelectionChanged()
    {
	int viewrow = table.getSelectedRow();
	if (viewrow<0)
	    return;

	int row = tableSorter.modelIndex(viewrow);

	if (row<0 || row > dupFileSets.size())
	    {
		//		pathList.setListData(new String[0]);
		return;
	    }

	FileSet fset = dupFileSets.get(row);

	String[] paths = new String[fset.files.size()];
	for (int i=0; i<fset.files.size();i++)
	    {
		paths[i]=fset.files.get(i).getAbsolutePath();
	    }
	pathList.setListData(paths);
    }

    public static void main(String[] args)
    {
	JUniq app = new JUniq();

	for (String s:args)
	    {
		try {
		    File f = new File(s);
		    if (f.isDirectory())
			app.rootQueue.put(new File(s));
		    else
			app.loadState(s);
		} catch (InterruptedException ex) {
		}
	    }
    }

    public void saveState(String path)
    {
	//	System.out.println("Writing state...");
	statusLabels.get(0).setText("Saving...");

	try {
	    FileOutputStream outs = new FileOutputStream(path);
	    ObjectOutputStream objouts = new ObjectOutputStream(new BufferedOutputStream(outs, 32768));
	    objouts.writeObject(digestTable);
	    objouts.writeObject(stats);
	    objouts.writeObject(symLinks);
	    objouts.writeObject(nonFiles);
	    objouts.close();
	} catch (IOException ex) {
	    System.out.println("Couldn't save database " + ex);
	}

	statusLabels.get(0).setText("idle");
    }

    public void loadState(String path)
    {
	statusLabels.get(0).setText("Loading...");

	try {
	    FileInputStream ins = new FileInputStream(path);
	    ObjectInputStream objins = new ObjectInputStream(new BufferedInputStream(ins, 32768));
	    digestTable = (HashMap<Long, ArrayList<FileSet>>) objins.readObject();
	    stats = (Stats) objins.readObject();
	    symLinks = (ArrayList<File>) objins.readObject();
	    nonFiles = (ArrayList<File>) objins.readObject();
	    objins.close();
	} catch (Exception ex) {
	    System.out.println("Couldn't read database " + ex);
	}

	for (ArrayList<FileSet> fsets : digestTable.values())
	    {
		for (FileSet fset : fsets)
		    {
			allFileSets.add(fset);
			if (fset.files.size()>1)
			    dupFileSets.add(fset);
		    }
	    }
	tableModel.fireTableDataChanged();
					
	updateStatus();

	statusLabels.get(0).setText("idle");
    }

    public static MessageDigest makeDigest()
    {
	MessageDigest md;

	try {
	    md = MessageDigest.getInstance("SHA1");
	} catch (NoSuchAlgorithmException ex) {
	    System.out.println("Couldn't create digest.");
	    System.exit(0);
	    return null;
	}

	return md;
    }

    public void scanFilesRecursive(File file, int rankin)
    {
    }

    void updateStatus()
    {
	for (int i=0; i<processThreads.size();i++)
	    {
		File file = processThreads.get(i).file;
		if (file != null)
		    statusLabels.get(i).setText("Processing : "+file.toString());
		else
		    statusLabels.get(i).setText("idle");
	    }

	locatedFilesLabel.setText(String.format("Located: %d files (%.1f MB)", stats.filesLocated,
						(stats.filesLocatedSize/1024/1024.0)));
	processedFilesLabel.setText(String.format("Processed: %d files (%.1f MB)", stats.filesScanned,
						  (stats.bytesTotal/1024/1024.0)));
	//	processedMbLabel.setText(String.format("Processed: %.1f MB", (stats.bytesTotal/1024/1024.0)));
	readMbLabel.setText(String.format("I/O: %.1f MB", (stats.bytesRead/1024/1024.0)));
	filesetsLabel.setText("Sets (dups/all): "+dupFileSets.size()+"/"+allFileSets.size());

	discardMbLabel.setText(String.format("Discardable: %d files (%.1f MB)",stats.filesDiscarded,
					     (stats.bytesDiscarded/1024/1024.0)));
    }

    static long byteArrayToLong(byte[] b)
    {
	long v = b[0] ^ (b[1]<<8) ^ (b[2]<<16) ^ (b[3]<<24) ^ (b[4]<<32) 
	    ^ (b[5]<<40) ^ (b[6]<<48) ^ (b[7]<<56);

	return v;
    }

    /*
      static long getFastDigest(File f)
      {
      return f.length();
      }

    */
    long getFastDigest(File f)
    {
	MessageDigest md = makeDigest();

	try {
	    FileInputStream ins = new FileInputStream(f);
	    byte[] buf = new byte[FASTHASH_LENGTH];
	    int len = ins.read(buf);
	    md.update(buf);
	    
	    long fastDigest = byteArrayToLong(md.digest());

	    ins.close();

	    stats.bytesRead += len;

	    return fastDigest ^ f.length();

	} catch (IOException ex) {
	    System.out.println("Couldn't process file " + f + " : " + ex);
	    return 0;
	}
    }

    byte[] getFullDigest(File f)
    {
	MessageDigest md = makeDigest();

	try {
	    FileInputStream ins = new FileInputStream(f);
	    byte[] buf = new byte[READBUF_LENGTH];

	    int len;
	    while ((len = ins.read(buf)) > 0)
		{
		    md.update(buf, 0, len);
		}

	    ins.close();

	    stats.bytesRead += f.length();

	    return md.digest();

	} catch (IOException ex) {
	    System.out.println("Couldn't process file " + f + " : " + ex);
	    return null;
	}
    }

    void ensureFullDigest(FileSet f)
    {
	if (f.fullDigest == null)
	    f.fullDigest = getFullDigest(f.files.get(0));
    }

    static boolean isSymbolicLink(File f)
    {
	try {
	    return (!f.getAbsolutePath().equals(f.getCanonicalPath()));
	} catch (IOException ex) {
	    System.out.println("Couldn't determine whether "+f+" is sym link! "+ex);
	    System.exit(0);
	    return false;
	}
    }

    void showActions()
    {
	/*
	for (FileSet fset: allFileSets)
	    {
		if (fset.discardfiles == null)
		    continue;

		textarea.append("--------\n");
		textarea.append("KEEP: "+fset.files.get(0)+"\n");
		textarea.append("DISC: ");

		boolean firsttime = true;
		for (File f: fset.discardfiles)
		    {
			if (!firsttime)
			    textarea.append("      ");

			textarea.append(""+f+"\n");

			firsttime = false;
		    }
	    }
	*/
    }

    boolean askToCloseCurrent()
    {
	if (currentDbPath != null)
	    {
		System.out.println("Throwing away current db!");
		currentDbPath = null;
		// if they abort, return false.

		return true;
	    }
	return true;
    }

    void fileNew()
    {
	if (!askToCloseCurrent())
	    return;

	digestTable = new HashMap<Long,ArrayList<FileSet>>();
	allFileSets = new ArrayList<FileSet>();
	dupFileSets = new ArrayList<FileSet>();
	symLinks = new ArrayList<File>();

	stats = new Stats();

	tableModel.fireTableDataChanged();

	updateStatus();
    }

    void fileOpen()
    {
	if (!askToCloseCurrent())
	    return;

	int ret = dbChooser.showOpenDialog(frame);
	if (ret != JFileChooser.APPROVE_OPTION)
	    return;

	loadState(dbChooser.getSelectedFile().getAbsolutePath());
    }

    void fileSave()
    {
	if (currentDbPath==null)
	    {
		fileSaveAs();
		return;
	    }

	saveState(currentDbPath);
    }

    void fileSaveAs()
    {
	int ret = dbChooser.showSaveDialog(frame);
	if (ret != JFileChooser.APPROVE_OPTION)
	    return;

	currentDbPath = dbChooser.getSelectedFile().getAbsolutePath();

	fileSave();
    }

    void fileQuit()
    {
	if (!askToCloseCurrent())
	    return;

	System.exit(0);
    }

    void opAdd()
    {
	pathChooser.setMultiSelectionEnabled(true);
	pathChooser.setFileSelectionMode(JFileChooser.FILES_AND_DIRECTORIES);

	int ret = pathChooser.showOpenDialog(frame);
	if (ret != JFileChooser.APPROVE_OPTION)
	    return;

	File[] files = pathChooser.getSelectedFiles();

	for (File f : files)
	    {
		try {
		    rootQueue.put(f);
		} catch (InterruptedException ex) {
		}
	    }
    }

    void opGenerate()
    {
	generate.setVisible(true);
    }

    class DirectorySearcherThread extends Thread
    {
	public DirectorySearcherThread()
	{
	    setDaemon(true);
	}

	public void run()
	{
	    while (true)
		{
		    try {
			File f = rootQueue.take();
			recurse(f);
		    } catch (InterruptedException ex) {
		    }
		}
	}

	public void recurse(File file)
	{
	    // ignore symlinks
	    if (isSymbolicLink(file))
		{
		    symLinks.add(file);
		    System.out.println("Detected symlink: "+file);
		    return;
		}

	    // descend recursively.
	    if (file.isDirectory())
		{
		    File[] files = file.listFiles();
		    if (files==null)
			return;

		    for (File thisfile : files)
			{
			    recurse(thisfile);
			}
		    return;
		}
	    
	    if (!file.isFile())
		{
		    System.out.println("Detected non file, non directory: " + file);
		    nonFiles.add(file);
		    return;
		}

	    // just enqueue this one, call it a day.
	    try {
		fileQueue.put(file);
	    } catch (InterruptedException ex) {
	    }

	    stats.filesLocated++;
	    stats.filesLocatedSize+=file.length();
	}
    }

    class StatusThread extends Thread
    {
	public StatusThread()
	{
	    setDaemon(true);
	}

	public void run()
	{
	    while (true)
		{
		    updateStatus();

		    try {
			Thread.sleep(200);
		    } catch (InterruptedException ex) {
		    }
		}
	}
    }

    class ProcessThread extends Thread
    {
	File file = null;

	public ProcessThread()
	{
	    setDaemon(true);
	    setPriority(MIN_PRIORITY);
	}

	public void run()
	{
	    while (true)
		{
		    try {
			file = fileQueue.take();
			process();
			file = null;
		    } catch (InterruptedException ex) {
		    }
		}
	}

	void process()
	{
	    long fileLength = file.length();
	    stats.bytesTotal += fileLength;

	    // process one file
	    stats.filesScanned++;

	    long fastDigest = getFastDigest(file);
	    if (fastDigest == 0)
		return;

	    ArrayList<FileSet> fsets = digestTable.get(fastDigest);
	    if (fsets == null)
		{
		    fsets = new ArrayList<FileSet>();
		    FileSet fset = new FileSet();
		    fset.fastDigest = fastDigest;
		    fset.files.add(file);
		    fset.fileLength = file.length();

		    fsets.add(fset);
		    digestTable.put(fastDigest, fsets);

		    allFileSets.add(fset);
		}
	    else
		{
		    boolean found = false;

		    // we found one or more matching file sets.
		    // before we do any expensive work, make sure that
		    // we haven't seen this exact file before
		    // (according to path name, i.e., they added the
		    // same path twice.)
		    for (FileSet fset : fsets)
			{
			    for (File f : fset.files)
				{
				    if (f.getAbsolutePath().equals(file.getAbsolutePath()))
					return;
				}
			}

		    // now we're going to need this. Bummer.
		    byte[] fullDigest = getFullDigest(file);
		
		    // okay, it's not a duplicate file. We need to resolve the collision.
		    for (FileSet fset : fsets)
			{
			    // make sure we've computed a digest for this fileset.
			    ensureFullDigest(fset);

			    // we found an identical file.
			    if (Arrays.equals(fset.fullDigest, fullDigest))
				{
				    fset.files.add(file);

				    if (fset.files.size()==2)
					{
					    dupFileSets.add(fset);
					    int idx = dupFileSets.size() - 1;
					    
					    fset.dupSetIndex = idx;
					    tableModel.fireTableRowsInserted(idx, idx);
					}

				    // we're going to discard something...
				    stats.bytesDiscarded += fileLength;
				    stats.filesDiscarded++;

				    found = true;

				    if (fset.dupSetIndex>=0)
					tableModel.fireTableRowsUpdated(fset.dupSetIndex, fset.dupSetIndex);

				    break;
				}
			}

		    // if no match, create a new file set.
		    if (!found)
			{
			    FileSet fset = new FileSet();
			    fset.fastDigest = fastDigest;
			    fset.fullDigest = fullDigest;
			    fset.files.add(file);
			    fset.fileLength = file.length();

			    fsets.add(fset);
			    allFileSets.add(fset);
			}
		}

	}
    }

    class UniqTableModel extends AbstractTableModel
    {
	public static final long serialVersionUID = 1001;

	public synchronized int getRowCount()
	{
	    return dupFileSets.size();
	}

	public int getColumnCount()
	{
	    return 4;
	}

	public Class getColumnClass(int column)
	{
	    if (column==0)
		return String.class;

	    return Long.class;
	}

	public Object getValueAt(int row, int column)
	{
	    FileSet fset = dupFileSets.get(row);

	    if (column==0)
		return fset.files.get(0).getAbsolutePath();
	    
	    /*
	    if (column==1)
		{
		    String[] vs=new String[fset.files.size()-1];

		    //		    String v="";
		    for (int i=1;i<fset.files.size();i++)
			{
			    vs[i-1]=fset.files.get(i).getAbsolutePath();
   //			    v=v+fset.files.get(i).getAbsolutePath()+"\n";
			}
		    
		    return new JList(vs);
		}
	    */

	    if (column==2)
		{
		    return fset.fileLength;
		    /*
		    if (fset.fileLength>1024*1024)
			return String.format("%.1f MB",fset.fileLength/1024/1024.0);
		    if (fset.fileLength>1024)
			return String.format("%.1f KB",fset.fileLength/1024.0);
		    return String.format("%d B", fset.fileLength);
		    */
		}

	    if (column==1)
		return fset.files.size()-1;

	    if (column==3)
		return (fset.files.size()-1)*fset.fileLength;
	    return "";
	}

	public String getColumnName(int col)
	{
	    switch (col)
		{
		case 0:
		    return "Exemplar File";
		case 2:
		    return "File size";
		case 1:
		    return "Dups";
		case 3:
		    return "Dups Size";
		}
	    return "?";
	}
    }

}
